import torch
from torch import nn, LongTensor
import torch.nn.functional as F
from fastai.core import split_kwargs_by_func
from fastprogress.fastprogress import progress_bar
from fastai.basic_train import Callback
from fastai.torch_core import add_metrics
from .._utils.pointcloud_od import confusion_matrix3d

import numpy as np
import random
import math


def conv_params(in_size, out_size):
    filters = [3, 2, 5, 4]
    strides = [1, 2, 3]  # max_stride = 3
    pads = [0, 1, 2, 3]  # max pad

    if out_size == 1:
        return 1, 0, in_size

    for filter_size in filters:
        for pad in pads:
            for stride in strides:
                if (out_size - 1) * stride == (in_size - filter_size) + 2 * pad:
                    return stride, pad, filter_size
    return None, None, None


def conv_paramsv2(in_size, out_size):
    filters = [3]
    strides = [1, 2]  # max_stride = 2
    pads = [0, 1]  # max pad

    if out_size == 1:
        return 1, 0, in_size

    for filter_size in filters:
        for pad in pads:
            for stride in strides:
                if (((in_size - filter_size) + 2 * pad) // stride) + 1 == out_size:
                    return stride, pad, filter_size
    return None, None, None


class StdConv(nn.Module):
    def __init__(self, nin, nout, filter_size=3, stride=2, padding=1, drop=0.1):
        super().__init__()
        self.conv = nn.Conv2d(nin, nout, filter_size, stride=stride, padding=padding)
        self.bn = nn.BatchNorm2d(nout)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        return self.drop(self.bn(F.relu(self.conv(x))))


class StdConvv2(nn.Module):
    def __init__(
        self,
        nin,
        nout,
        upsample_size=0,
        upsample=False,
        filter_size=3,
        stride=2,
        padding=1,
        drop=0.1,
    ):
        super().__init__()
        self.upsample = upsample
        self.conv = nn.Conv2d(nin, nout, filter_size, stride=stride, padding=padding)
        self.up = nn.Upsample(upsample_size)
        self.bn = nn.BatchNorm2d(nout)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        if self.upsample == True:
            return self.drop(self.bn(F.relu(self.conv(self.up(x)))))
        return self.drop(self.bn(F.relu(self.conv(x))))


def flatten_conv(x, k):
    bs, nf, gx, gy = x.size()
    x = x.permute(0, 2, 3, 1).contiguous()
    return x.view(bs, -1, nf // k)


class OutConv(nn.Module):
    def __init__(self, k, nin, num_classes, bias):
        super().__init__()
        self.k = k
        self.oconv1 = nn.Conv2d(nin, (num_classes) * k, 3, padding=1)
        self.oconv2 = nn.Conv2d(nin, 4 * k, 3, padding=1)
        self.oconv1.bias.data.zero_().add_(bias)

    def forward(self, x):
        return [
            flatten_conv(self.oconv1(x), self.k),
            flatten_conv(self.oconv2(x), self.k),
        ]


class SSDHead(nn.Module):
    def __init__(
        self,
        grids,
        anchors_per_cell,
        num_classes,
        num_features=7,
        drop=0.3,
        bias=-4.0,
        num_channels=512,
    ):
        super().__init__()
        self.drop = nn.Dropout(drop)

        self.sconvs = nn.ModuleList([])
        self.oconvs = nn.ModuleList([])

        self.anc_grids = grids

        self._k = anchors_per_cell

        self.sconvs.append(StdConv(num_channels, 256, stride=1, drop=drop))

        for i in range(len(grids)):
            if i == 0:
                stride, pad, filter_size = conv_params(num_features, grids[i])
            else:
                stride, pad, filter_size = conv_params(grids[i - 1], grids[i])

            if stride is None:
                print(grids[i - 1], " --> ", grids[i])
                raise Exception("cannot create model for specified grids.")

            self.sconvs.append(
                StdConv(256, 256, filter_size, stride=stride, padding=pad, drop=drop)
            )
            self.oconvs.append(
                OutConv(self._k, 256, num_classes=num_classes, bias=bias)
            )

    def forward(self, x):
        x = self.drop(F.relu(x))
        x = self.sconvs[0](x)
        out_classes = []
        out_bboxes = []
        for sconv, oconv in zip(self.sconvs[1:], self.oconvs):
            x = sconv(x)
            out_class, out_bbox = oconv(x)
            out_classes.append(out_class)
            out_bboxes.append(out_bbox)

        return [torch.cat(out_classes, dim=1), torch.cat(out_bboxes, dim=1)]


class SSDHeadv2(nn.Module):
    def __init__(
        self,
        grids,
        anchors_per_cell,
        num_classes,
        num_features=7,
        drop=0.3,
        bias=-4.0,
        num_channels=512,
    ):
        super().__init__()
        self.drop = nn.Dropout(drop)

        self.sconvs = nn.ModuleList([])
        self.oconvs = nn.ModuleList([])

        self.anc_grids = grids

        self._k = anchors_per_cell

        self.sconvs.append(StdConvv2(num_channels, 256, stride=1, drop=drop))

        for i in range(len(grids)):
            upsample = False

            if i == 0 and num_features >= grids[i]:
                stride, pad, filter_size = conv_paramsv2(num_features, grids[i])
                if stride is None:
                    upsample = True
                    stride, pad, filter_size = 1, 1, 3
            elif i == 0 and num_features < grids[i]:
                upsample = True
                stride, pad, filter_size = 1, 1, 3

            elif i != 0 and grids[i - 1] > grids[i]:
                stride, pad, filter_size = conv_paramsv2(grids[i - 1], grids[i])
                if stride is None:
                    upsample = True
                    stride, pad, filter_size = 1, 1, 3
            else:
                upsample = True
                stride, pad, filter_size = 1, 1, 3

            self.sconvs.append(
                StdConvv2(
                    256,
                    256,
                    grids[i],
                    upsample,
                    filter_size,
                    stride=stride,
                    padding=pad,
                    drop=drop,
                )
            )
            self.oconvs.append(
                OutConv(self._k, 256, num_classes=num_classes, bias=bias)
            )

    def forward(self, x):
        x = self.drop(F.relu(x))
        x = self.sconvs[0](x)
        out_classes = []
        out_bboxes = []
        for sconv, oconv in zip(self.sconvs[1:], self.oconvs):
            x = sconv(x)
            out_class, out_bbox = oconv(x)
            out_classes.append(out_class)
            out_bboxes.append(out_bbox)

        return [torch.cat(out_classes, dim=1), torch.cat(out_bboxes, dim=1)]


def one_hot_embedding(labels, num_classes):
    return torch.eye(num_classes)[labels.data.cpu()]


class BCE_Loss(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.num_classes = num_classes

    def forward(self, pred, targ):
        t = one_hot_embedding(targ, self.num_classes)
        t = torch.Tensor(t[:, 1:].contiguous()).to(pred.device)
        x = pred[:, 1:]
        w = self.get_weight(x, t)
        return F.binary_cross_entropy_with_logits(x, t, w, reduction="sum") / (
            self.num_classes - 1
        )

    def get_weight(self, x, t):
        return None


class FocalLoss(BCE_Loss):
    def get_weight(self, x, t):
        alpha, gamma = 0.25, 1
        p = x.sigmoid()
        pt = p * t + (1 - p) * (1 - t)
        w = alpha * t + (1 - alpha) * (1 - t)
        w = w * (1 - pt).pow(gamma)
        return w.detach()


def nms(boxes, scores, overlap=0.5, top_k=100):
    keep = scores.new(scores.size(0)).zero_().long()
    if boxes.numel() == 0:
        return keep
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    area = torch.mul(x2 - x1, y2 - y1)
    v, idx = scores.sort(0)  # sort in ascending order
    idx = idx[-top_k:]  # indices of the top-k largest vals
    xx1 = boxes.new()
    yy1 = boxes.new()
    xx2 = boxes.new()
    yy2 = boxes.new()
    w = boxes.new()
    h = boxes.new()

    count = 0
    while idx.numel() > 0:
        i = idx[-1]  # index of current largest val
        keep[count] = i
        count += 1
        if idx.size(0) == 1:
            break
        idx = idx[:-1]  # remove kept element from view
        # load bboxes of next highest vals
        torch.index_select(x1, 0, idx, out=xx1)
        torch.index_select(y1, 0, idx, out=yy1)
        torch.index_select(x2, 0, idx, out=xx2)
        torch.index_select(y2, 0, idx, out=yy2)
        # store element-wise max with next highest score
        xx1 = torch.clamp(xx1, min=x1[i])
        yy1 = torch.clamp(yy1, min=y1[i])
        xx2 = torch.clamp(xx2, max=x2[i])
        yy2 = torch.clamp(yy2, max=y2[i])
        w.resize_as_(xx2)
        h.resize_as_(yy2)
        w = xx2 - xx1
        h = yy2 - yy1
        # check sizes of xx1 and xx2.. after each iteration
        w = torch.clamp(w, min=0.0)
        h = torch.clamp(h, min=0.0)
        inter = w * h
        # IoU = i / (area(a) + area(b) - i)
        rem_areas = torch.index_select(area, 0, idx)  # load remaining areas)
        union = (rem_areas - inter) + area[i]
        IoU = inter / union  # store result in iou
        # keep only elements with an IoU <= overlap
        idx = idx[IoU.le(overlap)]
    return keep, count


def postprocess(
    pred, model=None, thresh=0.5, nms_overlap=0.1, ret_scores=True, device=None
):
    """
    It works on a single activation, does not support batch.
    """
    if device is None:
        device = torch.device("cpu")
    b_clas, b_bb = pred
    a_ic = model._actn_to_bb(
        b_bb.to(device), model._anchors.to(device), model._grid_sizes.to(device)
    )
    conf_scores, clas_ids = b_clas[:, 1:].max(1)
    conf_scores = b_clas.t().sigmoid().to(device)

    out1, bbox_list, class_list = [], [], []

    for cl in range(1, len(conf_scores)):
        c_mask = conf_scores[cl] > thresh
        if c_mask.sum() == 0:
            continue
        scores = conf_scores[cl][c_mask]
        l_mask = c_mask.unsqueeze(1)
        l_mask = l_mask.expand_as(a_ic)
        boxes = a_ic[l_mask].view(-1, 4)  # boxes are now in range[ 0, 1]
        boxes = (boxes - 0.5) * 2.0  # putting boxes in range[-1, 1]
        ids, count = nms(
            boxes.data, scores, nms_overlap, 50
        )  # FIX- NMS overlap hardcoded
        ids = ids[:count]
        out1.append(scores[ids])
        bbox_list.append(boxes.data[ids])
        class_list.append(torch.tensor([cl] * count))

    if len(bbox_list) == 0:
        return None

    if ret_scores:
        return (
            torch.cat(bbox_list, dim=0).to(device),
            torch.cat(class_list, dim=0).to(device),
            torch.cat(out1, dim=0).to(device),
        )
    else:
        return torch.cat(bbox_list, dim=0), torch.cat(class_list, dim=0)


class AveragePrecision(Callback):
    def __init__(self, model, n_classes, mode_3d=False):
        self.model = model
        self.n_classes = n_classes
        self.mode_3d = mode_3d

    def on_epoch_begin(self, **kwargs):
        self.tps, self.clas, self.p_scores = [], [], []
        self.classes, self.n_gts = (
            LongTensor(range(self.n_classes)),
            torch.zeros(self.n_classes).long(),
        )

    def on_batch_end(self, last_output, last_target, **kwargs):
        if (
            getattr(self.model, "_is_fasterrcnn", False)
            or "MMDetection" in self.model.__str__()
        ):
            last_output = last_output[0]

        if self.mode_3d:
            tps, p_scores, clas, self.n_gts = confusion_matrix3d(
                last_output, last_target, self.n_gts, self.classes
            )
        else:
            tps, p_scores, clas, self.n_gts = compute_cm(
                self.model, last_output, last_target, self.n_gts, self.classes
            )
        self.tps.extend(tps)
        self.p_scores.extend(p_scores)
        self.clas.extend(clas)

    def on_epoch_end(self, last_metrics, **kwargs):
        aps = compute_ap_score(
            self.tps, self.p_scores, self.clas, self.n_gts, self.n_classes, self.mode_3d
        )
        aps = torch.mean(torch.tensor(aps))
        return add_metrics(last_metrics, aps)


def compute_class_AP(
    model,
    dl,
    n_classes,
    show_progress,
    iou_thresh=0.5,
    detect_thresh=0.35,
    num_keep=100,
    **kwargs
):
    tps, clas, p_scores = [], [], []
    if getattr(model, "_is_model_extension", False):
        transform_kwargs, kwargs = split_kwargs_by_func(
            kwargs, model._model_conf.transform_input
        )
    classes, n_gts = LongTensor(range(n_classes)), torch.zeros(n_classes).long()
    with torch.no_grad():
        for input, target in progress_bar(dl, display=show_progress):
            if getattr(model, "_is_model_extension", False):
                try:
                    if model._is_multispectral:
                        output = model.learn.model.eval()(
                            model._model_conf.transform_input_multispectral(
                                input, **transform_kwargs
                            )
                        )
                    else:
                        output = model.learn.model.eval()(
                            model._model_conf.transform_input(input, **transform_kwargs)
                        )
                except Exception as e:
                    if getattr(model, "_is_fasterrcnn", False):
                        output = []
                        for _ in range(input.shape[0]):
                            res = {}
                            res["boxes"] = torch.empty(0, 4)
                            res["scores"] = torch.tensor([])
                            res["labels"] = torch.tensor([])
                            output.append(res)
                    else:
                        raise e
            else:
                output = model.learn.pred_batch(batch=(input, target))
            tps1, p_scores1, clas1, n_gts = compute_cm(
                model, output, target, n_gts, classes, iou_thresh, detect_thresh
            )
            tps.extend(tps1)
            p_scores.extend(p_scores1)
            clas.extend(clas1)

    aps = compute_ap_score(tps, p_scores, clas, n_gts, n_classes)
    return aps


def compute_cm(
    model, output, target, n_gts, classes, iou_thresh=0.1, detect_thresh=0.2
):
    tps, clas, p_scores = [], [], []

    if getattr(model, "_is_model_extension", False):
        analyzed_pred_out = model._analyze_pred(
            output,
            thresh=detect_thresh,
            nms_overlap=iou_thresh,
            ret_scores=True,
            device=model._device,
        )

    for i in range(target[0].size(0)):
        if getattr(model, "_is_model_extension", False):
            op = analyzed_pred_out[i]
        else:
            op = model._data.y.analyze_pred(
                (output[0][i], output[1][i]),
                model=model,
                thresh=detect_thresh,
                nms_overlap=iou_thresh,
                ret_scores=True,
                device=model._device,
            )
        tgt_bbox, tgt_clas = model._get_y(target[0][i], target[1][i])

        try:
            bbox_pred, preds, scores = op
            if len(bbox_pred) != 0 and len(tgt_bbox) != 0:
                ious = model._jaccard(bbox_pred, tgt_bbox)
                max_iou, matches = ious.max(1)
                detected = []
                for i in range(len(preds)):
                    if (
                        max_iou[i] >= iou_thresh
                        and matches[i] not in detected
                        and tgt_clas[matches[i]] == preds[i]
                    ):
                        detected.append(matches[i])
                        tps.append(1)
                    else:
                        tps.append(0)
                clas.append(preds.cpu())
                p_scores.append(scores.cpu())
        except Exception as e:
            pass
        n_gts += ((tgt_clas.cpu()[:, None] - 1) == classes[None, :]).sum(0)

    return tps, p_scores, clas, n_gts


def compute_ap_score(tps, p_scores, clas, n_gts, n_classes, mode_3d=False):
    # If no true positives are found return an average precision score of 0.
    if len(tps) == 0:
        return [0.0 for _ in range(n_classes)]

    tps, p_scores, clas = torch.tensor(tps), torch.cat(p_scores, 0), torch.cat(clas, 0)
    fps = 1 - tps
    idx = p_scores.argsort(descending=True)
    tps, fps, clas = tps[idx], fps[idx], clas[idx]
    aps = []
    for n_gts_idx, cls in enumerate(range(1, n_classes + 1)):
        if mode_3d:
            cls -= 1
        tps_cls, fps_cls = (
            tps[clas == cls].float().cumsum(0),
            fps[clas == cls].float().cumsum(0),
        )
        if tps_cls.numel() != 0 and tps_cls[-1] != 0:
            precision = tps_cls / (tps_cls + fps_cls + 1e-8)
            recall = tps_cls / (n_gts[n_gts_idx] + 1e-8)
            aps.append(compute_ap(precision, recall))
        else:
            aps.append(0.0)
    return aps


def compute_ap(precision, recall):
    "Compute the average precision for `precision` and `recall` curve."
    recall = np.concatenate(([0.0], list(recall), [1.0]))
    precision = np.concatenate(([0.0], list(precision), [0.0]))
    for i in range(len(precision) - 1, 0, -1):
        precision[i - 1] = np.maximum(precision[i - 1], precision[i])
    idx = np.where(recall[1:] != recall[:-1])[0]
    ap = np.sum((recall[idx + 1] - recall[idx]) * precision[idx + 1])
    return ap


def iou(ann, centroids):
    similarities = []

    for centroid in centroids:
        inter = np.prod(np.minimum(ann, centroid))
        union = np.prod(ann) + np.prod(centroid) - inter
        similarities.append(inter / union)

    return np.array(similarities)


def avg_iou(bboxes, centroids):
    sum = 0.0

    for bbox in bboxes:
        sum += max(iou(bbox, centroids))

    return sum / bboxes.shape[0]


def kmeans(bboxes, num_anchor):
    # Method based on https://github.com/experiencor/keras-yolo3

    num_points, dim = bboxes.shape
    prev_centroids = np.ones(num_points) * (-1)
    indices = [random.randrange(num_points) for i in range(num_anchor)]
    centroids = bboxes[indices]

    while True:
        distances = []
        for bbox in bboxes:
            d = 1 - iou(bbox, centroids)
            distances.append(d)
        distances = np.array(distances)
        cur_centroids = np.argmin(distances, axis=1)

        if (prev_centroids == cur_centroids).all():
            return centroids

        centroid_sums = np.zeros(
            (num_points, dim), float
        )  # num_points needs to be num_anchors
        for i in range(num_points):
            centroid_sums[cur_centroids[i]] += bboxes[i]
        for i in range(num_anchor):
            centroids[i] = centroid_sums[i] / (np.sum(cur_centroids == i) + 1e-6)

        prev_centroids = cur_centroids.copy()



import json
import traceback
import logging

try:
    from ._ner_spacy import _SpacyEntityRecognizer
    from .._utils._ner_utils import spaCyNERDatabunch

    HAS_SPACY = True
except Exception as e:
    spacy_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_SPACY = False

try:
    from ._ner_transformer import _TransformerEntityRecognizer
    from .._utils.text_data import TextDataObject
    from .._utils.common import _get_emd_path
    from transformers import AutoConfig

    HAS_TRANSFORMERS = True
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMERS = False

    class _TransformerEntityRecognizer:
        supported_backbones = []


def _raise_spacy_import_error():
    error_message = (
        f"{spacy_exception}\n\n\n"
        "This module requires spacy version 2.1.8 or above and fastprogress."
        "Install it using 'pip install spacy==2.1.8 fastprogress pandas'"
    )
    raise Exception(error_message)


def _raise_transformers_import_error():
    error_message = (
        f"{transformer_exception}\n\n\n"
        "This module requires transformers version 3.3.0, "
        "install it using 'pip install transformers==3.3.0'"
    )
    raise Exception(error_message)


class EntityRecognizer:
    """
    Creates an entity recognition model to extract text entities from unstructured text documents.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Optional data object returned from :meth:`~arcgis.learn.prepare_data` function.
                            data object can be `None`, in case where someone wants to use a
                            Hugging Face Transformer model fine-tuned on entity-recognition
                            task. In this case the model should be used directly for inference.
    ---------------------   -------------------------------------------
    lang                    Optional string. Language-specific code,
                            named according to the language’s `ISO code <https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes>`_
                            The default value is 'en' for English.
    ---------------------   -------------------------------------------
    backbone                Optional string. Specify `spacy` or the  HuggingFace
                            transformer model name to be used to train the
                            entity recognizer model. Default set to `spacy`.

                            Entity recognition via `spaCy` is based on <https://spacy.io/api/entityrecognizer>

                            To learn more about the available transformer models or
                            choose models that are suitable for your dataset,
                            kindly visit:- https://huggingface.co/transformers/pretrained_models.html

                            To learn more about the available transformer models fine-tuned
                            on Named Entity Recognition Task, kindly visit:-
                            https://huggingface.co/models?pipeline_tag=token-classification
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    verbose                 Optional string. Default set to `error`. The
                            log level you want to set. It means the amount
                            of information you want to display while training
                            or calling the various methods of this class.
                            Allowed values are - `debug`, `info`, `warning`,
                            `error` and `critical`. Applicable only for models
                            with HuggingFace transformer backbones.
    ---------------------   -------------------------------------------
    seq_len                 Optional Integer. Default set to 512. Maximum
                            sequence length (at sub-word level after tokenization)
                            of the training data to be considered for training
                            the model. Applicable only for models with
                            HuggingFace transformer backbones.
    ---------------------   -------------------------------------------
    mixed_precision         Optional Bool. Default set to False. If set
                            True, then mixed precision training is used
                            to train the model. Applicable only for models
                            with HuggingFace transformer backbones.
    ---------------------   -------------------------------------------
    pretrained_path         Optional String. Path where pre-trained model
                            is saved. Accepts a Deep Learning Package
                            (DLPK) or Esri Model Definition(EMD) file.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.text.EntityRecognizer` Object
    """

    supported_backbones = ["spacy"] + _TransformerEntityRecognizer.supported_backbones

    def __init__(self, data, lang="en", backbone="spacy", **kwargs):
        self.data = data
        self.lang = lang
        self.backbone = backbone
        self.entities = None
        create_empty = kwargs.get("create_empty", False)

        if create_empty:
            pass
        else:
            if data is None:
                model = EntityRecognizer.from_pretrained(backbone, **kwargs)
                self._model = model._model
                self.entities = self._model.entities
            elif data and backbone == "spacy":
                if not HAS_SPACY:
                    _raise_spacy_import_error()
                if data.backbone != "spacy":
                    logging.info("Preparing data for spacy backbone!")
                    data.prepare_data_for_spacy()
                data_obj = data.get_data_object()
                self._model = _SpacyEntityRecognizer(data_obj, lang=lang, **kwargs)
            elif data:
                if not HAS_TRANSFORMERS:
                    _raise_transformers_import_error()
                model_config = AutoConfig.from_pretrained(backbone)

                if data.backbone == "spacy":
                    logging.info("Preparing data for transformer backbone!")
                    if model_config.id2label != {0: "LABEL_0", 1: "LABEL_1"}:
                        label2id, id2label = (
                            model_config.label2id,
                            model_config.id2label,
                        )
                        # sci-bert has `label2id` mapping as {"LABEL_0": 0, "LABEL_1": 1, "LABEL_10": 10...} and
                        # `id2label` mapping as {"0": "I-cell_type", "1": "B-DNA", "2": "O"...}, hence this hack
                        if label2id != {y: x for x, y in id2label.items()} and not any(
                            ["-" in x for x in label2id.keys()]
                        ):
                            label2id = {y: x for x, y in id2label.items()}
                        labels = list(label2id.keys())
                    else:
                        labels, label2id = [], None
                    logging.info(f"Labels - {labels}\n\tlabel2id mappings - {label2id}")
                    ignore_tag_order = not any(["-" in label for label in labels])
                    data.prepare_data_for_transformer(
                        ignore_tag_order=ignore_tag_order, label2id=label2id
                    )
                data_obj = data.get_data_object()
                kwargs.update({"model_config": model_config})
                self._model = _TransformerEntityRecognizer(data_obj, backbone, **kwargs)

        if create_empty is False:
            self.train_ds = self._model.train_ds
            self.valid_ds = self._model.val_ds

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "precision_score", "recall_score", "f1_score"]

    @property
    def available_metrics(self):
        """
        List of available metrics that are displayed in the training
        table. Set `monitor` value to be one of these while calling
        the `fit` method.
        """
        return ["valid_loss", "precision_score", "recall_score", "f1_score"]

    @classmethod
    def available_backbone_models(cls, architecture):
        """
        Get available models for the given entity recognition backbone

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        architecture            Required string. name of the architecture
                                one wishes to use. To learn more about
                                the available models or choose models that are
                                suitable for your dataset, kindly visit:-
                                https://huggingface.co/transformers/pretrained_models.html
        =====================   ===========================================

        :return: a tuple containing the available models for the given entity recognition backbone
        """

        if architecture == "spacy":
            return ("spacy",)
        else:
            return _TransformerEntityRecognizer.available_backbone_models(architecture)

    def lr_find(self, allow_plot=True):
        """
        Runs the Learning Rate Finder. Helps in choosing the
        optimum learning rate for training the model.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        allow_plot              Optional boolean. Display the plot of losses
                                against the learning rates and mark the optimal
                                value of the learning rate on the plot.
                                The default value is 'True'.
        =====================   ===========================================
        """
        return self._model.lr_find(allow_plot=allow_plot)

    def unfreeze(self):
        """
        Unfreezes the earlier layers of the model for fine-tuning.
        """
        self._model.unfreeze()

    def freeze(self):
        """
        Freeze up to last layer group to train only the last layer group of the model.
        """
        self._model.freeze()

    def fit(
        self,
        epochs=20,
        lr=None,
        one_cycle=True,
        early_stopping=False,
        checkpoint=True,
        **kwargs,
    ):
        """
        Train the model for the specified number of epochs and using the
        specified learning rates

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Required integer. Number of cycles of training
                                on the data. Increase it if underfitting.
        ---------------------   -------------------------------------------
        lr                      Optional float or slice of floats. Learning rate
                                to be used for training the model. If ``lr=None``,
                                an optimal learning rate is automatically deduced
                                for training the model.

                                .. note::
                                    Passing slice of floats as `lr` value is not supported for models with `spaCy` backbone.
        ---------------------   -------------------------------------------
        one_cycle               Optional boolean. Parameter to select 1cycle
                                learning rate schedule. If set to `False` no
                                learning rate schedule is used.

                                .. note::
                                        Not applicable for models with spaCy backbone
        ---------------------   -------------------------------------------
        early_stopping          Optional boolean. Parameter to add early stopping.
                                If set to 'True' training will stop if parameter
                                `monitor` value stops improving for 5 epochs.

                                .. note::
                                    Not applicable for models with spaCy backbone
        ---------------------   -------------------------------------------
        checkpoint              Optional boolean or string.
                                Parameter to save checkpoint during training.
                                If set to `True` the best model
                                based on `monitor` will be saved during
                                training. If set to 'all', all checkpoints
                                are saved. If set to False, checkpointing will
                                be off. Setting this parameter loads the best
                                model at the end of training.

                                .. note::
                                    Not applicable for models with spaCy backbone
        ---------------------   -------------------------------------------
        tensorboard             Optional boolean. Parameter to write the training log.
                                If set to 'True' the log will be saved at
                                <dataset-path>/training_log which can be visualized in
                                tensorboard. Required tensorboardx version=2.1

                                The default value is 'False'.

                                .. note::
                                    Not applicable for Text Models
        ---------------------   -------------------------------------------
        monitor                 Optional string. Parameter specifies
                                which metric to monitor while checkpointing
                                and early stopping. Defaults to 'valid_loss'. Value
                                should be one of the metric that is displayed in
                                the training table. Use `{model_name}.available_metrics`
                                to list the available metrics to set here.

                                .. note::
                                        Not applicable for models with spaCy backbone
        =====================   ===========================================
        """

        self._model.fit(
            epochs=epochs,
            lr=lr,
            one_cycle=one_cycle,
            early_stopping=early_stopping,
            checkpoint=checkpoint,
            **kwargs,
        )
        self.entities = self._model.entities

    def save(self, name_or_path, **kwargs):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to save. It
                                stores it at the pre-defined location. If path
                                is passed then it stores at the specified path
                                with model name as directory name and creates
                                all the intermediate directories.
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
                                Default is set to False.
        ---------------------   -------------------------------------------
        gis                     Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        compute_metrics         Optional boolean. Used for computing model
                                metrics. Default is set to True.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False
                                Not applicable for models with `spaCy` backbone.
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
                                Boolean `zip_files` if True, it will create the Deep
                                Learning Package (DLPK) file while saving the model.
        =====================   ===========================================
        """

        return self._model.save(name_or_path=name_or_path, **kwargs)

    def load(self, name_or_path):
        """
        Loads a saved EntityRecognizer model from disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        =====================   ===========================================
        """

        self._model.load(name_or_path=name_or_path)
        self.entities = self._model.entities

    @classmethod
    def from_pretrained(cls, backbone, **kwargs):
        """
        Creates an EntityRecognizer model object from an already fine-tuned
        Hugging Face Transformer backbone.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        backbone                Required string. Specify the Hugging Face Transformer
                                backbone name fine-tuned on Named Entity Recognition(NER)/
                                Token Classification task.

                                To get more details on available transformer models
                                fine-tuned on Named Entity Recognition(NER) Task, kindly visit:-
                                https://huggingface.co/models?pipeline_tag=token-classification
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        verbose                 Optional string. Default set to `error`. The
                                log level you want to set. It means the amount
                                of information you want to display while calling
                                the various methods of this class. Allowed values
                                are - `debug`, `info`, `warning`, `error` and `critical`.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.text.EntityRecognizer` Object
        """

        if "spacy" in backbone:
            error_message = (
                f"Wrong backbone - `{backbone}` supplied. Only HuggingFace model names fine-tuned on "
                "`TokenClassification` tasks are allowed to be passed as `backbone` in the method."
            )
            raise Exception(error_message)

        model_config = AutoConfig.from_pretrained(backbone)
        if model_config.id2label == {0: "LABEL_0", 1: "LABEL_1"}:
            error_message = (
                f"Wrong backbone - `{backbone}` supplied. This backbone is not fine-tuned on a"
                "`TokenClassification` task. Kindly choose an appropriate backbone to call this method."
                "Visit:- https://huggingface.co/models?pipeline_tag=token-classification to find models"
                " for `NamedEntityRecognition` or `TokenClassification` tasks."
            )
            raise Exception(error_message)

        label2id, id2label = model_config.label2id, model_config.id2label
        # sci-bert has `label2id` mapping as {"LABEL_0": 0, "LABEL_1": 1, "LABEL_10": 10...} and
        # `id2label` mapping as {"0": "I-cell_type", "1": "B-DNA", "2": "O"...}, hence this hack
        if label2id != {y: x for x, y in id2label.items()} and not any(
            ["-" in x for x in label2id.keys()]
        ):
            label2id = {y: x for x, y in id2label.items()}

        model = _TransformerEntityRecognizer._from_pretrained(
            backbone, label2id, **kwargs
        )

        clas_object = cls(data=None, backbone=backbone, create_empty=True)

        clas_object._model = model
        clas_object.entities = clas_object._model.entities
        clas_object.train_ds = clas_object._model.train_ds
        clas_object.valid_ds = clas_object._model.val_ds

        return clas_object

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates an EntityRecognizer model object from a Deep Learning
        Package(DLPK) or Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required DatabunchNER object or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.text.EntityRecognizer` Object
        """

        data_obj = None
        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd_json = json.load(f)
        backbone = emd_json.get("ModelType", "spacy").lower()
        if backbone == "spacy":
            if data and data.backbone != "spacy":
                logging.info("Preparing data for spacy backbone!")
                data.prepare_data_for_spacy()
            if data:
                data_obj = data.get_data_object()
            model = _SpacyEntityRecognizer.from_model(emd_path=emd_path, data=data_obj)
        else:
            if data and data.backbone == "spacy":
                logging.info("Preparing data for transformer backbone!")
                labels = set(emd_json["Labels"])
                label2id = emd_json["Label2Id"]
                logging.info(f"Labels - {labels}\n\tlabel2id mappings - {label2id}")
                ignore_tag_order = not any(["-" in label for label in labels])
                data.prepare_data_for_transformer(
                    ignore_tag_order=ignore_tag_order, label2id=label2id
                )
            if data:
                data_obj = data.get_data_object()
            model = _TransformerEntityRecognizer.from_model(
                emd_path=emd_path, data=data_obj
            )

        clas_object = cls(data=None, backbone=backbone, create_empty=True)

        clas_object._model = model
        clas_object.entities = clas_object._model.entities
        clas_object.train_ds = clas_object._model.train_ds
        clas_object.valid_ds = clas_object._model.val_ds

        return clas_object

    def extract_entities(self, text_list, drop=True, batch_size=4, show_progress=True):
        """
        Extracts the entities from [documents in the mentioned path or text_list].

        Field defined as 'address_tag' in :meth:`~arcgis.learn.prepare_data`  function's class mapping
        attribute will be treated as a location. In cases where trained model extracts
        multiple locations from a single document, that document will be replicated
        for each location in the resulting dataframe.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_list               Required string(path) or list(documents).
                                List of documents for entity extraction OR
                                path to the documents.
        ---------------------   -------------------------------------------
        drop                    Optional bool. If documents without address
                                needs to be dropped from the results.
                                Default is set to True.
        ---------------------   -------------------------------------------
        batch_size              Optional integer. Number of items to process
                                at once. (Reduce it if getting CUDA Out of Memory
                                Errors). Default is set to 4.
                                Not applicable for models with `spaCy` backbone.
        ---------------------   -------------------------------------------
        show_progress           optional Bool. If set to True, will display a
                                progress bar depicting the items processed so far.
                                Applicable only when a list of text is passed
        =====================   ===========================================

        :return: Pandas DataFrame
        """

        return self._model.extract_entities(
            text_list, drop=drop, batch_size=batch_size, show_progress=show_progress
        )

    def show_results(self, ds_type="valid"):
        """
        Runs entity extraction on a random batch from the mentioned ds_type.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        ds_type                 Optional string, defaults to valid.
        =====================   ===========================================

        :return: Pandas DataFrame
        """

        return self._model.show_results(ds_type=ds_type)

    def precision_score(self):
        """
        Calculate precision score of the trained model
        """
        return self._model.precision_score()

    def recall_score(self):
        """
        Calculate recall score of the trained model
        """
        return self._model.recall_score()

    def f1_score(self):
        """
        Calculate F1 score of the trained model
        """
        return self._model.f1_score()

    def metrics_per_label(self):
        """
        Calculate precision, recall & F1 scores per labels/entities
        for which the model was trained on
        """
        return self._model.metrics_per_label()

    def plot_losses(self, show=True):
        """
        Plot training and validation losses.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        show                    Optional bool. Defaults to True
                                If set to False, figure will not be plotted
                                but will be returned, when set to True function
                                will plot the figure and return nothing.
        =====================   ===========================================

        :return: `matplotlib.figure.Figure <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure>`_
        """

        return self._model.plot_losses(show=show)



from pathlib import Path
import json
import warnings
from ._model_extension import ModelExtension

try:
    from fastai.vision import flatten_model, ImageList
    from fastai.vision import imagenet_stats, normalize
    import torch
    from fastai.torch_core import split_model_idx
    from .._utils.pascal_voc_rectangles import ObjectDetectionCategoryList
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from ._ssd_utils import AveragePrecision

    HAS_FASTAI = True

except Exception as e:
    HAS_FASTAI = False


class MMDetectionConfig:
    try:
        import torch
        import types
        import numpy
        import os
        import pathlib
    except:
        pass

    def get_model(self, data, backbone=None, **kwargs):
        import mmdet.models
        import mmcv
        import logging

        logging.disable(logging.WARNING)

        config = kwargs.get("model", False)
        checkpoint = kwargs.get("model_weight", False)
        if config[-2:] != "py":
            config += ".py"
        if self.os.path.exists(self.pathlib.Path(config)):
            cfg = mmcv.Config.fromfile(config)
            cfg.model.pretrained = None
        else:
            import arcgis

            cfg_abs_path = (
                self.pathlib.Path(arcgis.__file__).parent
                / "learn"
                / "_mmdetection_config"
                / config
            )
            cfg = mmcv.Config.fromfile(cfg_abs_path)
            checkpoint = cfg.get("checkpoint", False)
            if checkpoint:
                cfg.model.pretrained = None

        if hasattr(cfg.model, "roi_head"):
            if isinstance(cfg.model.roi_head.bbox_head, list):
                for box_head in cfg.model.roi_head.bbox_head:
                    box_head.num_classes = data.c - 1
            else:
                cfg.model.roi_head.bbox_head.num_classes = data.c - 1
        else:
            cfg.model.bbox_head.num_classes = data.c - 1

        if cfg.model.backbone.type == "DetectoRS_ResNet" and getattr(
            data, "_is_multispectral", False
        ):
            if hasattr(cfg.model.neck, "rfp_backbone"):
                cfg.model.neck.rfp_backbone.in_channels = len(data._extract_bands)

        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model = mmdet.models.build_detector(cfg.model)

        if checkpoint:
            mmcv.runner.load_checkpoint(
                model, checkpoint, "cpu", False, logging.getLogger()
            )

        from mmcv.runner import auto_fp16

        @auto_fp16(apply_to=("img",))
        def forward_modified(self, img, img_metas=None, gt_bboxes=None, gt_labels=None):
            if self.training:
                losses = self.forward_train(img, img_metas, gt_bboxes, gt_labels)
                loss, log_vars = self._parse_losses(losses)

                loss = dict(loss=loss, log_vars=log_vars)

                output = None
                if self.train_val:
                    output = self.forward_test([img], [img_metas], rescale=True)
                return output, loss
            else:
                return self.forward_test(img[0], img[1], rescale=True)

        model.forward = self.types.MethodType(forward_modified, model)

        self.model = model
        self.cfg = cfg

        logging.disable(0)

        return model

    def on_batch_begin(self, learn, model_input_batch, model_target_batch, **kwargs):
        if kwargs.get("train"):
            self.model.train_val = False
        else:
            self.set_test_parms()
            self.model.train_val = True
        learn.model.train()
        img_metas = []
        gt_labels = []
        gt_bboxes = []
        image_pad_shape = model_input_batch.permute(0, 2, 3, 1).shape[1:]
        image_scale_factor = self.numpy.array(
            [1.0, 1.0, 1.0, 1.0], dtype=self.numpy.float32
        )

        for bboxes, classes in zip(*model_target_batch):
            non_pad_index = bboxes.sum(dim=1) != 0
            bboxes = bboxes[non_pad_index]
            classes = classes[non_pad_index] - 1

            bboxes = ((bboxes + 1) / 2) * learn.data.chip_size
            if bboxes.nelement() == 0:
                bboxes = self.torch.tensor([[0.0, 0.0, 1.0, 1.0]]).to(learn.data.device)
                classes = self.torch.tensor([0]).to(learn.data.device)

            bboxes = self.torch.index_select(
                bboxes, 1, self.torch.tensor([1, 0, 3, 2]).to(learn.data.device)
            )

            img_metas_dict = {}
            img_metas_dict["pad_shape"] = image_pad_shape
            img_metas_dict["img_shape"] = image_pad_shape
            img_metas_dict["ori_shape"] = image_pad_shape
            img_metas_dict["scale_factor"] = image_scale_factor

            gt_bboxes.append(bboxes)
            gt_labels.append(classes)
            img_metas.append(img_metas_dict)

        model_input = [model_input_batch, img_metas, gt_bboxes, gt_labels]
        # Model target is not required in traing mode so just return the same model_target to train the model.
        model_target = model_target_batch

        # return model_input and model_target
        return model_input, model_target

    def set_test_parms(self, thresh=0.2, nms_overlap=0.1):
        if hasattr(self.model, "roi_head"):
            self.nms_thres = self.model.roi_head.test_cfg.nms.iou_threshold
            self.thresh = self.model.roi_head.test_cfg.score_thr
            self.model.roi_head.test_cfg.nms.iou_threshold = nms_overlap
            self.model.roi_head.test_cfg.score_thr = thresh
        else:
            self.nms_thres = self.model.bbox_head.test_cfg.nms.iou_threshold
            self.thresh = self.model.bbox_head.test_cfg.score_thr
            self.model.bbox_head.test_cfg.nms.iou_threshold = nms_overlap
            self.model.bbox_head.test_cfg.score_thr = thresh

    def transform_input(self, xb, thresh=0.5, nms_overlap=0.1):
        self.set_test_parms(thresh, nms_overlap)
        img_metas = []
        image_pad_shape = xb.permute(0, 2, 3, 1).shape[1:]
        image_scale_factor = self.numpy.array(
            [1.0, 1.0, 1.0, 1.0], dtype=self.numpy.float32
        )

        for i in range(xb.shape[0]):
            img_metas_dict = {}
            img_metas_dict["pad_shape"] = image_pad_shape
            img_metas_dict["img_shape"] = image_pad_shape
            img_metas_dict["ori_shape"] = image_pad_shape
            img_metas_dict["scale_factor"] = image_scale_factor
            img_metas.append(img_metas_dict)

        model_input = [[xb], [img_metas]]
        return model_input

    def transform_input_multispectral(self, xb, thresh=0.5, nms_overlap=0.1):
        return self.transform_input(xb, thresh, nms_overlap)

    def loss(self, model_output, *model_target):
        return model_output[1]["loss"]

    def post_process(self, pred, nms_overlap, thres, chip_size, device):
        if hasattr(self.model, "roi_head"):
            self.model.roi_head.test_cfg.nms.iou_threshold = self.nms_thres
            thres = self.model.roi_head.test_cfg.score_thr
            self.model.roi_head.test_cfg.score_thr = self.thresh
        else:
            self.model.bbox_head.test_cfg.nms.iou_threshold = self.nms_thres
            thres = self.model.bbox_head.test_cfg.score_thr
            self.model.bbox_head.test_cfg.score_thr = self.thresh

        post_processed_pred = []
        for p in pred:
            bbox = self.numpy.vstack(p)
            label = [
                self.numpy.full(box.shape[0], i, dtype=self.numpy.int32)
                for i, box in enumerate(p)
            ]
            label = self.numpy.concatenate(label) + 1
            score = bbox[:, -1]
            bbox = bbox[:, 0:-1]
            kip_pred = score > thres
            bbox, label, score = (
                self.torch.from_numpy(bbox[kip_pred]),
                self.torch.from_numpy(label[kip_pred]),
                self.torch.from_numpy(score[kip_pred]),
            )
            # convert bboxes in range -1 to 1.
            bbox = bbox / (chip_size / 2) - 1
            # convert bboxes in format [y1,x1,y2,x2]
            bbox = self.torch.index_select(
                bbox, 1, self.torch.tensor([1, 0, 3, 2]).to(bbox.device)
            )
            # Append the tuple in list for each image
            post_processed_pred.append(
                (bbox.data.to(device), label.to(device), score.to(device))
            )

        return post_processed_pred


class MMDetection(ModelExtension):
    """
    =============================   =============================================
    **Parameter**                    **Description**
    -----------------------------   ---------------------------------------------
    data                            Required fastai Databunch. Returned data object from
                                    :meth:`~arcgis.learn.prepare_data`  function.
    -----------------------------   ---------------------------------------------
    model                           Required model name or path to the configuration file
                                    from :class:`~arcgis.learn.MMDetection` repository. The list of the
                                    supported models can be queried using
                                    :attr:`~arcgis.learn.MMDetection.supported_models` .
    -----------------------------   ---------------------------------------------
    model_weight                    Optional path of the model weight from
                                    :class:`~arcgis.learn.MMDetection` repository.
    -----------------------------   ---------------------------------------------
    pretrained_path                 Optional string. Path where pre-trained model is
                                    saved.
    =============================   =============================================

    :return: :class:`~arcgis.learn.MMDetection` Object
    """

    def __init__(self, data, model, model_weight=False, pretrained_path=None, **kwargs):
        self._check_dataset_support(data)

        super().__init__(
            data,
            MMDetectionConfig,
            pretrained_path=pretrained_path,
            model=model,
            model_weight=model_weight,
        )
        self.learn.metrics = [AveragePrecision(self, data.c - 1)]
        idx = self._freeze()
        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])
        self.learn.create_opt(lr=3e-3)

    def unfreeze(self):
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    def _freeze(self):
        "Freezes the pretrained backbone."
        for idx, i in enumerate(flatten_model(self.learn.model.backbone)):
            if isinstance(i, (torch.nn.BatchNorm2d)):
                continue
            for p in i.parameters():
                p.requires_grad = False
        return idx

    @staticmethod
    def _available_metrics():
        return ["valid_loss"]

    @property
    def _is_mmsegdet(self):
        return True

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return MMDetection._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["PASCAL_VOC_rectangles", "KITTI_rectangles"]

    supported_models = [
        "atss",
        "carafe",
        "cascade_rcnn",
        "cascade_rpn",
        "dcn",
        "detectors",
        "double_heads",
        "dynamic_rcnn",
        "empirical_attention",
        "fcos",
        "foveabox",
        "fsaf",
        "ghm",
        "hrnet",
        "libra_rcnn",
        "nas_fcos",
        "pafpn",
        "pisa",
        "regnet",
        "reppoints",
        "res2net",
        "sabl",
        "vfnet",
    ]
    """
    List of models supported by this class.
    """

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.MMDetection` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.MMDetection` Object
        """
        emd_path = _get_emd_path(emd_path)

        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        dataset_type = emd.get("DatasetType", "PASCAL_VOC_rectangles")
        chip_size = emd["ImageWidth"]
        resize_to = emd.get("resize_to", None)
        kwargs = emd.get("Kwargs", {})
        if isinstance(resize_to, list):
            resize_to = (resize_to[0], resize_to[1])

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        data_passed = True
        if data is None:
            data_passed = False
            train_tfms = []
            val_tfms = []
            ds_tfms = (train_tfms, val_tfms)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                sd = ImageList([], path=emd_path.parent.parent).split_by_idx([])
                data = (
                    sd.label_const(
                        0,
                        label_cls=ObjectDetectionCategoryList,
                        classes=list(class_mapping.values()),
                    )
                    .transform(ds_tfms)
                    .databunch()
                    .normalize(imagenet_stats)
                )
            # Add 1 for background class
            data.c += 1
            data.chip_size = chip_size
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data.classes = ["background"] + list(class_mapping.values())
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)
            data.dataset_type = dataset_type

        data.resize_to = resize_to
        frcnn = cls(data, pretrained_path=str(model_file), **kwargs)

        if not data_passed:
            frcnn.learn.data.single_ds.classes = frcnn._data.classes
            frcnn.learn.data.single_ds.y.classes = frcnn._data.classes

        return frcnn

    def predict(
        self,
        image_path,
        threshold=0.5,
        nms_overlap=0.1,
        return_scores=False,
        visualize=False,
        resize=False,
    ):
        """
        Runs prediction on an Image. This method is only supported for RGB images.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean. Will return the probability
                                scores of the bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with
                                predicted bounding boxes if True.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the same size
                                (chip_size parameter in prepare_data) that the model was trained on,
                                before detecting objects.
                                Note that if resize_to parameter was used in prepare_data,
                                the image is resized to that size instead.

                                By default, this parameter is false and the detections are run
                                in a sliding window fashion by applying the model on cropped sections
                                of the image (of the same size as the model was trained on).
        =====================   ===========================================

        :return: Returns a tuple with predictions, labels and optionally confidence scores
                  if return_scores=True. The predicted bounding boxes are returned as a list
                  of lists containing the  xmin, ymin, width and height of each predicted object
                  in each image. The labels are returned as a list of class values and the
                  confidence scores are returned as a list of floats indicating the confidence
                  of each prediction.
        """

    def predict_video(
        self,
        input_video_path,
        metadata_file,
        threshold=0.5,
        nms_overlap=0.1,
        track=False,
        visualize=False,
        output_file_path=None,
        multiplex=False,
        multiplex_file_path=None,
        tracker_options={
            "assignment_iou_thrd": 0.3,
            "vanish_frames": 40,
            "detect_frames": 10,
        },
        visual_options={
            "show_scores": True,
            "show_labels": True,
            "thickness": 2,
            "fontface": 0,
            "color": (255, 255, 255),
        },
        resize=False,
    ):
        """
        Runs prediction on a video and appends the output VMTI predictions in the metadata file.
        This method is only supported for RGB images.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        input_video_path        Required. Path to the video file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        metadata_file           Required. Path to the metadata csv file where
                                the predictions will be saved in VMTI format.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        track                   Optional bool. Set this parameter as True to
                                enable object tracking.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. If True a video is saved
                                with prediction results.
        ---------------------   -------------------------------------------
        output_file_path        Optional path. Path of the final video to be saved.
                                If not supplied, video will be saved at path input_video_path
                                appended with _prediction.
        ---------------------   -------------------------------------------
        multiplex               Optional boolean. Runs Multiplex using the VMTI detections.
        ---------------------   -------------------------------------------
        multiplex_file_path     Optional path. Path of the multiplexed video to be saved.
                                By default a new file with _multiplex.MOV extension is saved
                                in the same folder.
        ---------------------   -------------------------------------------
        tracking_options        Optional dictionary. Set different parameters for
                                object tracking. assignment_iou_thrd parameter is used
                                to assign threshold for assignment of trackers,
                                vanish_frames is the number of frames the object should
                                be absent to consider it as vanished, detect_frames
                                is the number of frames an object should be detected
                                to track it.
        ---------------------   -------------------------------------------
        visual_options          Optional dictionary. Set different parameters for
                                visualization.
                                show_scores boolean, to view scores on predictions,
                                show_labels boolean, to view labels on predictions,
                                thickness integer, to set the thickness level of box,
                                fontface integer, fontface value from opencv values,
                                color tuple (B, G, R), tuple containing values between
                                0-255.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the video frames to the same size
                                (chip_size parameter in prepare_data) that the model was trained on,
                                before detecting objects.
                                Note that if resize_to parameter was used in prepare_data,
                                the video frames are resized to that size instead.

                                By default, this parameter is false and the detections are run
                                in a sliding window fashion by applying the model on cropped sections
                                of the frame (of the same size as the model was trained on).
        =====================   ===========================================

        """

    def average_precision_score(
        self, detect_thresh=0.2, iou_thresh=0.1, mean=False, show_progress=True
    ):
        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """

    def show_results(self, rows=5, thresh=0.5, nms_overlap=0.1):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        =====================   ===========================================

        """



import random
from uuid import uuid4
import string
import os
import pandas as pd
import tempfile
import shutil
from arcgis._impl.common._utils import _date_handler
from arcgis.gis import Item, ItemDependency
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._isd import InsensitiveDict
from arcgis.auth.tools import LazyLoader

_tool_utils = LazyLoader("arcgis.features.geo._tools._utils")
_common_utils = LazyLoader("arcgis._impl.common._utils")
features = LazyLoader("arcgis.features")
json = LazyLoader("json")

try:
    from arcgis.features.geo import _is_geoenabled
except:

    def _is_geoenabled(o):
        return False


try:
    import arcpy

    has_arcpy = True
except ImportError:
    has_arcpy = False
except RuntimeError:
    has_arcpy = False
try:
    import shapefile

    has_pyshp = True
except ImportError:
    has_pyshp = False


def _json_encode_params(postdata):
    for k, v in postdata.items():
        if isinstance(v, (dict, list, tuple, bool)):
            postdata[k] = json.dumps(v, default=_date_handler)
        elif isinstance(v, PropertyMap):
            postdata[k] = json.dumps(dict(v), default=_date_handler)
        elif isinstance(v, InsensitiveDict):
            postdata[k] = v.json

    return postdata


def _create_file_item(gis, df, file_type, **kwargs):
    try:
        # File Type Dictionary
        ftypes = {"File Geodatabase": "gdb", "Shapefile": "shp", "CSV": "csv"}

        # Pop out kwargs, establish params to be used throughout
        service_name = kwargs.pop("service_name", None)
        if service_name is None:
            service_name = "a" + uuid4().hex[:7]
        temp_dir = os.path.join(tempfile.gettempdir(), service_name)
        title = kwargs.pop("title", uuid4().hex)
        capabilities = kwargs.pop("capabilities", "Query")
        item_id = kwargs.pop("item_id", None)
        tags = kwargs.pop("tags", file_type)
        folder = kwargs.pop("folder", None)
        name = "%s%s.%s" % (
            random.choice(string.ascii_lowercase),
            uuid4().hex[:5],
            ftypes[file_type],
        )

        # Create the file to be added as an item
        if file_type in ["File Geodatabase", "Shapefile"]:
            # Working with feature layers
            # set up temporary zip to be used in directory
            os.makedirs(temp_dir)
            temp_zip = os.path.join(temp_dir, "%s.zip" % ("a" + uuid4().hex[:5]))

            # Create filegdb or shapefile
            if file_type == "File Geodatabase":
                # create empty filegdb
                emtpy_fgdb = _tool_utils.run_and_hide(
                    fn=arcpy.CreateFileGDB_management,
                    **{"out_folder_path": temp_dir, "out_name": name},
                )
                fgdb = emtpy_fgdb[0]
                location = os.path.join(fgdb, os.path.basename(temp_dir))
                zip_loc = os.path.join(temp_dir, name)
            else:
                location = os.path.join(temp_dir, name)
                zip_loc = temp_dir

            # Writes the df to file as features
            sanitize_columns = kwargs.pop("sanitize_columns", False)
            df.spatial.to_featureclass(
                location=location, sanitize_columns=sanitize_columns
            )

            # zip it
            file = _common_utils.zipws(path=zip_loc, outfile=temp_zip, keep=True)

        elif file_type == "CSV":
            # Table Workflow
            file = tempfile.gettempdir() + "\\%s%s.csv" % (
                random.choice(string.ascii_lowercase),
                uuid4().hex[:5],
            )
            with open(file, "w") as my_csv:
                df.to_csv(my_csv)
                my_csv.close()

        # add item to portal
        file_item = gis.content.add(
            item_properties={
                "title": title,
                "type": file_type,
                "tags": tags,
            },
            data=file,
            folder=folder,
        )

        if file_type == "CSV":
            # analyze the csv for publish params
            publish_parameters = gis.content.analyze(item=file_item, file_type="csv")
            publish_parameters["name"] = service_name
            publish_parameters["locationType"] = None
        else:
            # start creating publish params from new file item
            publish_parameters = {
                "hasStaticData": True,
                "name": os.path.splitext(file_item["name"])[0],
                "maxRecordCount": 2000,
                "layerInfo": {"capabilities": capabilities},
                "targetSR": kwargs.pop("target_sr", 102100),
            }

        new_item = file_item.publish(
            publish_parameters=publish_parameters, item_id=item_id
        )
    finally:
        # Clean up temporary files and directories
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir, ignore_errors=True)

        if os.path.exists(file):
            os.remove(file)
    return file_item, new_item


def _perform_overwrite(fl_index, flc_manager, layer_definition):
    # update the name and id to represent correct values
    layer_definition["id"] = fl_index
    layer_definition["name"] = flc_manager.properties.layers[fl_index]["name"]

    # Perform edit on the flc
    # Step 1: Preserve layer ids
    revert = False
    if (
        "preserveLayerIds" not in flc_manager.properties
        or flc_manager.properties["preserveLayerIds"] is not True
    ):
        flc_manager.update_definition({"preserveLayerIds": True})
        revert = True
    # Step 2: Delete layer from definition
    flc_manager.delete_from_definition({"layers": [{"id": fl_index}]})
    # Step 3: Add new layer to definition
    flc_manager.add_to_definition({"layers": [dict(layer_definition)]})
    # Step 4: Cleanup
    if revert:
        flc_manager.update_definition({"preserveLayerIds": False})


def _perform_insert(flc_manager, layer_definition):
    # Add new layer to definition
    flc_manager.add_to_definition({"layers": [dict(layer_definition)]})
    # Find the index at which the layer was added
    for layer in flc_manager.properties.layers:
        if layer["name"] == layer_definition["name"]:
            fl_index = layer["id"]
    return fl_index


def _add_item_dependency(
    file_type, fl_index, file_item, fs_item, new_item=None, gis=None
):
    if file_type.lower() == "csv":
        source_info = gis.content.analyze(item=file_item)["publishParameters"]
        ItemDependency(fs_item).add("itemid", file_item.id)
        fs_item.tables[fl_index].append(
            item_id=file_item.id,
            upload_format=file_type.lower(),
            source_info=source_info,
        )
    elif file_type.lower() == "shapefile" or (
        len(fs_item.layers) > 0
        and "filegdb" in fs_item.layers[fl_index].properties.supportedAppendFormats
    ):
        # correct file type for append method
        if file_type.lower() == "file geodatabase":
            file_type = "filegdb"
        else:
            file_type = "shapefile"
        ItemDependency(fs_item).add("itemid", file_item.id)
        fs_item.layers[fl_index].append(item_id=file_item.id, upload_format=file_type)
    else:
        # When filegdb not supported through append, use featureCollection
        features = new_item.layers[0].query().features
        fs_item.layers[fl_index].edit_features(adds=features)
    fs_item.add_relationship(rel_item=file_item, rel_type="Service2Data")


def import_as_item(gis, df, **kwargs):
    # House Keeping
    overwrite = kwargs.pop("overwrite", False)
    insert = kwargs.pop("append", False)

    if isinstance(df, features.FeatureSet):
        df = df.sdf

    # Check whether it will be a layer or a table
    if _is_geoenabled(df):
        # layer
        if has_arcpy == False and has_pyshp == False:
            raise Exception(
                "Spatially enabled DataFrame's must have either pyshp or"
                + " arcpy available to use import_data"
            )
        if has_arcpy:
            file_type = "File Geodatabase"
        elif has_pyshp:
            file_type = "Shapefile"
    else:
        # table
        file_type = "CSV"

    # Create the file item, new item published from the file item, and the publish parameters
    file_item, new_item = _create_file_item(gis, df, file_type, **kwargs)

    # If not overwrite or insert, return the new item
    if not (overwrite or insert):
        return new_item
    else:
        # Get user defined parameters to continue the workflow and either overwrite or insert
        fs_dict = kwargs.pop("service", None)
        if fs_dict is None:
            raise ValueError(
                "If overwite or append is True, then the feature service id needs to be specified in the `service` parameter."
            )

        # Get the fs_id and make sure correct format
        fs_id = fs_dict["featureServiceId"]
        if fs_id is None:
            raise ValueError(
                "The provided feature service id cannot be found. Please check it is correct and try again."
            )
        elif isinstance(fs_id, Item):
            fs_id = fs_id.itemid

        # Index passed in for overwrite, None for insert
        # If None, it will be assigned in the _perform_insert method
        index = fs_dict["layer"]

        # Create the feature layer manager for the existing feature service
        fs_item = gis.content.get(fs_id)
        flc_manager = features.FeatureLayerCollection.fromitem(fs_item).manager

        if len(new_item.layers) > 0:
            layer_definition = new_item.layers[0].properties
        elif len(new_item.tables) > 0:
            layer_definition = new_item.tables[0].properties

    if overwrite:
        # overwrite workflow
        _perform_overwrite(index, flc_manager, layer_definition)
    else:
        # insert workflow
        index = _perform_insert(flc_manager, layer_definition)

    # This pushes the features and adds new dependencies
    _add_item_dependency(file_type, index, file_item, fs_item, new_item, gis)

    # clean up
    new_item.delete()

    return fs_item


def import_as_fc(gis, df, **kwargs):
    # Get kwargs
    address_fields = kwargs.pop("address_fields", None)
    item_id = kwargs.pop("item_id", None)

    # Step 1: Analyze the df as a csv
    if kwargs.get("geocode_url", None):
        geocode_url = kwargs.get("geocode_url")
    else:
        locators = [
            gc["url"]
            for gc in gis.properties.helperServices.geocode
            if gc.get("batch", False)
        ]
        if len(locators) == 0:
            raise Exception("No batch geocoding service found.")
        geocode_url = locators[0]

    path = gis._public_rest_url + "content/features/analyze"

    postdata = {
        "f": "json",
        "text": df.to_csv(index=False),
        "filetype": "csv",
        "analyzeParameters": {
            "enableGlobalGeocoding": "true",
            "sourceLocale": kwargs.pop("source_locale", "us-en"),
            "sourceCountry": kwargs.pop("source_country", ""),
            "sourceCountryHint": kwargs.pop("country_hint", ""),
            "geocodeServiceUrl": geocode_url,
        },
    }
    if address_fields is not None:
        postdata["analyzeParameters"]["locationType"] = "address"

    postdata = _json_encode_params(postdata)
    resp = gis._con._session.post(url=path, data=postdata, timeout=600)
    res = resp.json()

    # Step 2: Prep parameters to generate features
    if address_fields is not None:
        res["publishParameters"].update({"addressFields": address_fields})
    path = gis._public_rest_url + "content/features/generate"
    postdata = {
        "f": "json",
        "text": df.to_csv(),
        "filetype": "csv",
        "publishParameters": json.dumps(res["publishParameters"]),
    }
    if item_id:
        postdata["itemIdToCreate"] = item_id

    if isinstance(df, pd.DataFrame) and "location_type" not in kwargs:
        # Step 2: Generate features
        postdata = _json_encode_params(postdata)
        resp = gis._con._session.post(path, postdata)
        res_generate = resp.json()
    elif (isinstance(df, pd.DataFrame) and "location_type" in kwargs) or (
        isinstance(df, pd.DataFrame) and address_fields
    ):
        # Step 2: Generate features
        if address_fields is not None:
            res["publishParameters"].update({"addressFields": address_fields})

        update_dict = {}
        update_dict["locationType"] = kwargs.pop("location_type", "")
        update_dict["latitudeFieldName"] = kwargs.pop("latitude_field", "")
        update_dict["longitudeFieldName"] = kwargs.pop("longitude_field", "")
        update_dict["coordinateFieldName"] = kwargs.pop("coordinate_field_name", "")
        update_dict["coordinateFieldType"] = kwargs.pop("coordinate_field_type", "")
        rk = []
        for k, v in update_dict.items():
            if v == "":
                rk.append(k)
        for k in rk:
            del update_dict[k]
        res["publishParameters"].update(update_dict)

        postdata = _json_encode_params(postdata)
        resp = gis._con._session.post(
            path, postdata
        )  # , use_ordered_dict=True) - OrderedDict >36< _mixins.PropertyMap

        res_generate = resp.json()

    # Step 3: Return
    if res_generate:
        return features.FeatureCollection(
            res_generate["featureCollection"]["layers"][0]
        )
    else:
        return None
